@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1997-2015. All Rights Reserved.
+%% Copyright Ericsson AB 1997-2018. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -34,9 +34,11 @@
 	 ip/1, stats/0, options/0, 
 	 pushf/3, popf/1, close/1, gethostname/0, gethostname/1, 
 	 parse_ipv4_address/1, parse_ipv6_address/1, parse_ipv4strict_address/1,
-	 parse_ipv6strict_address/1, parse_address/1, parse_strict_address/1, ntoa/1]).
+	 parse_ipv6strict_address/1, parse_address/1, parse_strict_address/1,
+         ntoa/1, ipv4_mapped_ipv6_address/1]).
 
 -export([connect_options/2, listen_options/2, udp_options/2, sctp_options/2]).
+-export([udp_module/1, tcp_module/1, tcp_module/2, sctp_module/1]).
 
 -export([i/0, i/1, i/2]).
 
@@ -71,10 +73,11 @@
 %% timer interface
 -export([start_timer/1, timeout/1, timeout/2, stop_timer/1]).
 
--export_type([address_family/0, hostent/0, hostname/0, ip4_address/0,
-              ip6_address/0, ip_address/0, posix/0, socket/0,
-              port_number/0]).
-
+-export_type([address_family/0, socket_protocol/0, hostent/0, hostname/0, ip4_address/0,
+              ip6_address/0, ip_address/0, port_number/0,
+	      local_address/0, socket_address/0, returned_non_ip_address/0,
+	      socket_setopt/0, socket_getopt/0,
+	      posix/0, socket/0, stat_option/0]).
 %% imports
 -import(lists, [append/1, duplicate/2, filter/2, foldl/3]).
 
@@ -97,7 +100,25 @@
 			0..65535,0..65535,0..65535,0..65535}.
 -type ip_address() :: ip4_address() | ip6_address().
 -type port_number() :: 0..65535.
--type posix() :: exbadport | exbadseq | file:posix().
+-type local_address() :: {local, File :: binary() | string()}.
+-type returned_non_ip_address() ::
+	{local, binary()} |
+	{unspec, <<>>} |
+	{undefined, any()}.
+-type posix() ::
+        'eaddrinuse' | 'eaddrnotavail' | 'eafnosupport' | 'ealready' |
+        'econnaborted' | 'econnrefused' | 'econnreset' |
+        'edestaddrreq' |
+        'ehostdown' | 'ehostunreach' |
+        'einprogress' | 'eisconn' |
+        'emsgsize' |
+        'enetdown' | 'enetunreach' |
+        'enopkg' | 'enoprotoopt' | 'enotconn' | 'enotty' | 'enotsock' |
+        'eproto' | 'eprotonosupport' | 'eprototype' |
+        'esocktnosupport' |
+        'etimedout' |
+        'ewouldblock' |
+        'exbadport' | 'exbadseq' | file:posix().
 -type socket() :: port().
 
 -type socket_setopt() ::
@@ -133,16 +154,19 @@
 		 'running' | 'multicast' | 'loopback']} |
       {'hwaddr', ether_address()}.
 
--type address_family() :: 'inet' | 'inet6'.
+-type address_family() :: 'inet' | 'inet6' | 'local'.
 -type socket_protocol() :: 'tcp' | 'udp' | 'sctp'.
 -type socket_type() :: 'stream' | 'dgram' | 'seqpacket'.
+-type socket_address() ::
+	ip_address() | 'any' | 'loopback' | local_address().
 -type stat_option() :: 
 	'recv_cnt' | 'recv_max' | 'recv_avg' | 'recv_oct' | 'recv_dvi' |
 	'send_cnt' | 'send_max' | 'send_avg' | 'send_oct' | 'send_pend'.
 
 %%% ---------------------------------
 
--spec get_rc() -> [{Par :: any(), Val :: any()}].
+-spec get_rc() -> [{Par :: atom(), Val :: any()} |
+                   {Par :: atom(), Val1 :: any(), Val2 :: any()}].
 
 get_rc() ->
     inet_db:get_rc().
@@ -160,26 +184,33 @@ close(Socket) ->
     end.
 
 
--spec peername(Socket) ->  {ok, {Address, Port}} | {error, posix()} when
-      Socket :: socket(),
-      Address :: ip_address(),
-      Port :: non_neg_integer().
+-spec peername(Socket :: socket()) ->
+		      {ok,
+		       {ip_address(), port_number()} |
+		       returned_non_ip_address()} |
+		      {error, posix()}.
 
 peername(Socket) -> 
     prim_inet:peername(Socket).
 
--spec setpeername(Socket :: socket(), Address :: {ip_address(), port_number()}) ->
-	'ok' | {'error', any()}.
+-spec setpeername(
+	Socket :: socket(),
+	Address ::
+	  {ip_address() | 'any' | 'loopback',
+	   port_number()} |
+	  socket_address()) ->
+			 'ok' | {'error', any()}.
 
 setpeername(Socket, {IP,Port}) ->
     prim_inet:setpeername(Socket, {IP,Port});
 setpeername(Socket, undefined) ->
     prim_inet:setpeername(Socket, undefined).
 
--spec peernames(Socket) -> {ok, [{Address, Port}]} | {error, posix()} when
-      Socket :: socket(),
-      Address :: ip_address(),
-      Port :: non_neg_integer().
+-spec peernames(Socket :: socket()) ->
+		       {ok,
+			[{ip_address(), port_number()} |
+			 returned_non_ip_address()]} |
+		       {error, posix()}.
 
 peernames(Socket) ->
     prim_inet:peernames(Socket).
@@ -195,15 +226,21 @@ peernames(Socket, Assoc) ->
     prim_inet:peernames(Socket, Assoc).
 
 
--spec sockname(Socket) -> {ok, {Address, Port}} | {error, posix()} when
-      Socket :: socket(),
-      Address :: ip_address(),
-      Port :: non_neg_integer().
+-spec sockname(Socket :: socket()) ->
+		      {ok,
+		       {ip_address(), port_number()} |
+		       returned_non_ip_address()} |
+		      {error, posix()}.
 
 sockname(Socket) -> 
     prim_inet:sockname(Socket).
 
--spec setsockname(Socket :: socket(), Address :: {ip_address(), port_number()}) ->
+-spec setsockname(
+	Socket :: socket(),
+	Address ::
+	  {ip_address() | 'any' | 'loopback',
+	   port_number()} |
+	  socket_address()) ->
 	'ok' | {'error', any()}.
 
 setsockname(Socket, {IP,Port}) -> 
@@ -211,10 +248,11 @@ setsockname(Socket, {IP,Port}) ->
 setsockname(Socket, undefined) ->
     prim_inet:setsockname(Socket, undefined).
 
--spec socknames(Socket) -> {ok, [{Address, Port}]} | {error, posix()} when
-      Socket :: socket(),
-      Address :: ip_address(),
-      Port :: non_neg_integer().
+-spec socknames(Socket :: socket()) ->
+		       {ok,
+			[{ip_address(), port_number()} |
+			 returned_non_ip_address()]} |
+		       {error, posix()}.
 
 socknames(Socket) ->
     prim_inet:socknames(Socket).
@@ -289,7 +327,7 @@ getifaddrs(Socket) ->
 -spec getifaddrs() -> {ok, Iflist} | {error, posix()} when
       Iflist :: [{Ifname,[Ifopt]}],
       Ifname :: string(),
-      Ifopt :: {flag,[Flag]} | {addr,Addr} | {netmask,Netmask}
+      Ifopt :: {flags,[Flag]} | {addr,Addr} | {netmask,Netmask}
              | {broadaddr,Broadaddr} | {dstaddr,Dstaddr}
              | {hwaddr,Hwaddr},
       Flag :: up | broadcast | loopback | pointtopoint
@@ -439,7 +477,12 @@ getstat(Socket,What) ->
       Hostent :: hostent().
 
 gethostbyname(Name) -> 
-    gethostbyname_tm(Name, inet, false).
+    case inet_db:res_option(inet6) of
+	true ->
+	    gethostbyname_tm(Name, inet6, false);
+	false ->
+	    gethostbyname_tm(Name, inet, false)
+    end.
 
 -spec gethostbyname(Hostname, Family) ->
                            {ok, Hostent} | {error, posix()} when
@@ -646,10 +689,18 @@ parse_address(Addr) ->
 parse_strict_address(Addr) ->
     inet_parse:strict_address(Addr).
 
+-spec ipv4_mapped_ipv6_address(ip_address()) -> ip_address().
+ipv4_mapped_ipv6_address({D1,D2,D3,D4})
+  when (D1 bor D2 bor D3 bor D4) < 256 ->
+    {0,0,0,0,0,16#ffff,(D1 bsl 8) bor D2,(D3 bsl 8) bor D4};
+ipv4_mapped_ipv6_address({D1,D2,D3,D4,D5,D6,D7,D8})
+  when (D1 bor D2 bor D3 bor D4 bor D5 bor D6 bor D7 bor D8) < 65536 ->
+    {D7 bsr 8,D7 band 255,D8 bsr 8,D8 band 255}.
+
 %% Return a list of available options
 options() ->
     [
-     tos, priority, reuseaddr, keepalive, dontroute, linger,
+     tos, tclass, priority, reuseaddr, keepalive, dontroute, linger,
      broadcast, sndbuf, recbuf, nodelay, ipv6_v6only,
      buffer, header, active, packet, deliver, mode,
      multicast_if, multicast_ttl, multicast_loop,
@@ -670,13 +721,13 @@ stats() ->
 %% Available options for tcp:connect
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 connect_options() ->
-    [tos, priority, reuseaddr, keepalive, linger, sndbuf, recbuf, nodelay,
+    [tos, tclass, priority, reuseaddr, keepalive, linger, sndbuf, recbuf, nodelay,
      header, active, packet, packet_size, buffer, mode, deliver, line_delimiter,
      exit_on_close, high_watermark, low_watermark, high_msgq_watermark,
      low_msgq_watermark, send_timeout, send_timeout_close, delay_send, raw,
-     show_econnreset].
+     show_econnreset, bind_to_device].
     
-connect_options(Opts, Family) ->
+connect_options(Opts, Mod) ->
     BaseOpts = 
 	case application:get_env(kernel, inet_default_connect_options) of
 	    {ok,List} when is_list(List) ->
@@ -693,7 +744,7 @@ connect_options(Opts, Family) ->
 	{ok, R} ->
 	    {ok, R#connect_opts {
 		   opts = lists:reverse(R#connect_opts.opts),
-		   ifaddr = translate_ip(R#connect_opts.ifaddr, Family)
+		   ifaddr = Mod:translate_ip(R#connect_opts.ifaddr)
 		  }};
 	Error -> Error	    
     end.
@@ -708,9 +759,6 @@ con_opt([Opt | Opts], #connect_opts{} = R, As) ->
 	{fd,Fd}     -> con_opt(Opts, R#connect_opts { fd = Fd }, As);
 	binary      -> con_add(mode, binary, R, Opts, As);
 	list        -> con_add(mode, list, R, Opts, As);
-	{tcp_module,_}  -> con_opt(Opts, R, As);
-	inet        -> con_opt(Opts, R, As);
-	inet6       -> con_opt(Opts, R, As);
 	{netns,NS} ->
 	    BinNS = filename2binary(NS),
 	    case prim_inet:is_sockopt_val(netns, BinNS) of
@@ -741,13 +789,13 @@ con_add(Name, Val, #connect_opts{} = R, Opts, AllOpts) ->
 %% Available options for tcp:listen
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 listen_options() ->
-    [tos, priority, reuseaddr, keepalive, linger, sndbuf, recbuf, nodelay,
+    [tos, tclass, priority, reuseaddr, keepalive, linger, sndbuf, recbuf, nodelay,
      header, active, packet, buffer, mode, deliver, backlog, ipv6_v6only,
      exit_on_close, high_watermark, low_watermark, high_msgq_watermark,
      low_msgq_watermark, send_timeout, send_timeout_close, delay_send,
-     packet_size, raw, show_econnreset].
+     packet_size, raw, show_econnreset, bind_to_device].
 
-listen_options(Opts, Family) ->
+listen_options(Opts, Mod) ->
     BaseOpts = 
 	case application:get_env(kernel, inet_default_listen_options) of
 	    {ok,List} when is_list(List) ->
@@ -764,7 +812,7 @@ listen_options(Opts, Family) ->
 	{ok, R} ->
 	    {ok, R#listen_opts {
 		   opts = lists:reverse(R#listen_opts.opts),
-		   ifaddr = translate_ip(R#listen_opts.ifaddr, Family)
+		   ifaddr = Mod:translate_ip(R#listen_opts.ifaddr)
 		  }};
 	Error -> Error
     end.
@@ -780,9 +828,6 @@ list_opt([Opt | Opts], #listen_opts{} = R, As) ->
 	{backlog,BL} ->  list_opt(Opts, R#listen_opts { backlog = BL }, As);
 	binary       ->  list_add(mode, binary, R, Opts, As);
 	list         ->  list_add(mode, list, R, Opts, As);
-	{tcp_module,_}  -> list_opt(Opts, R, As);
-	inet         -> list_opt(Opts, R, As);
-	inet6        -> list_opt(Opts, R, As);
 	{netns,NS} ->
 	    BinNS = filename2binary(NS),
 	    case prim_inet:is_sockopt_val(netns, BinNS) of
@@ -807,23 +852,36 @@ list_add(Name, Val, #listen_opts{} = R, Opts, As) ->
 	Error -> Error
     end.
 
+tcp_module(Opts) ->
+    tcp_module_1(Opts, undefined).
+
+tcp_module(Opts, Addr) ->
+    Address = {undefined,Addr},
+    %% Address has to be a 2-tuple but the first element is ignored
+    tcp_module_1(Opts, Address).
+
+tcp_module_1(Opts, Address) ->
+    mod(
+      Opts, tcp_module, Address,
+      #{inet => inet_tcp, inet6 => inet6_tcp, local => local_tcp}).
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% Available options for udp:open
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 udp_options() ->
-    [tos, priority, reuseaddr, sndbuf, recbuf, header, active, buffer, mode, 
+    [tos, tclass, priority, reuseaddr, sndbuf, recbuf, header, active, buffer, mode,
      deliver, ipv6_v6only,
      broadcast, dontroute, multicast_if, multicast_ttl, multicast_loop,
      add_membership, drop_membership, read_packets,raw,
-     high_msgq_watermark, low_msgq_watermark].
+     high_msgq_watermark, low_msgq_watermark, bind_to_device].
 
 
-udp_options(Opts, Family) ->
+udp_options(Opts, Mod) ->
     case udp_opt(Opts, #udp_opts { }, udp_options()) of
 	{ok, R} ->
 	    {ok, R#udp_opts {
 		   opts = lists:reverse(R#udp_opts.opts),
-		   ifaddr = translate_ip(R#udp_opts.ifaddr, Family)
+		   ifaddr = Mod:translate_ip(R#udp_opts.ifaddr)
 		  }};
 	Error -> Error
     end.
@@ -838,9 +896,6 @@ udp_opt([Opt | Opts], #udp_opts{} = R, As) ->
 	{fd,Fd}     ->  udp_opt(Opts, R#udp_opts { fd = Fd }, As);
 	binary      ->  udp_add(mode, binary, R, Opts, As);
 	list        ->  udp_add(mode, list, R, Opts, As);
-	{udp_module,_} -> udp_opt(Opts, R, As);
-	inet        -> udp_opt(Opts, R, As);
-	inet6       -> udp_opt(Opts, R, As);
 	{netns,NS} ->
 	    BinNS = filename2binary(NS),
 	    case prim_inet:is_sockopt_val(netns, BinNS) of
@@ -865,6 +920,11 @@ udp_add(Name, Val, #udp_opts{} = R, Opts, As) ->
 	Error -> Error
     end.
 
+udp_module(Opts) ->
+    mod(
+      Opts, udp_module, undefined,
+      #{inet => inet_udp, inet6 => inet6_udp, local => local_udp}).
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% Available options for sctp:open
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -880,8 +940,9 @@ udp_add(Name, Val, #udp_opts{} = R, Opts, As) ->
 %  (*) passing of open FDs ("fdopen") is not supported.
 sctp_options() ->
 [   % The following are generic inet options supported for SCTP sockets:
-    mode, active, buffer, tos, priority, dontroute, reuseaddr, linger, sndbuf,
+    mode, active, buffer, tos, tclass, priority, dontroute, reuseaddr, linger, sndbuf,
     recbuf, ipv6_v6only, high_msgq_watermark, low_msgq_watermark,
+    bind_to_device,
 
     % Other options are SCTP-specific (though they may be similar to their
     % TCP and UDP counter-parts):
@@ -921,9 +982,6 @@ sctp_opt([Opt|Opts], Mod, #sctp_opts{} = R, As) ->
 	    sctp_opt(Opts, Mod, R#sctp_opts{type=Type}, As);
 	binary		-> sctp_opt (Opts, Mod, R, As, mode, binary);
 	list		-> sctp_opt (Opts, Mod, R, As, mode, list);
-	{sctp_module,_}	-> sctp_opt (Opts, Mod, R, As); % Done with
-	inet		-> sctp_opt (Opts, Mod, R, As); % Done with
-	inet6		-> sctp_opt (Opts, Mod, R, As); % Done with
 	{netns,NS} ->
 	    BinNS = filename2binary(NS),
 	    case prim_inet:is_sockopt_val(netns, BinNS) of
@@ -965,6 +1023,11 @@ sctp_opt_ifaddr(Opts, Mod, #sctp_opts{ifaddr=IfAddr}=R, As, Addr) ->
 			  _                      -> [IP,IfAddr]
 		      end}, As).
 
+sctp_module(Opts) ->
+    mod(
+      Opts, sctp_module, undefined,
+      #{inet => inet_sctp, inet6 => inet6_sctp}).
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% Util to check and insert option in option list
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -1016,13 +1079,59 @@ binary2filename(Bin) ->
 	    Bin
     end.
 
-
 translate_ip(any,      inet) -> {0,0,0,0};
 translate_ip(loopback, inet) -> {127,0,0,1};
 translate_ip(any,      inet6) -> {0,0,0,0,0,0,0,0};
 translate_ip(loopback, inet6) -> {0,0,0,0,0,0,0,1};
 translate_ip(IP, _) -> IP.
 
+mod(Opts, Tag, Address, Map) ->
+    mod(Opts, Tag, Address, Map, undefined, []).
+%%
+mod([{Tag, M}|Opts], Tag, Address, Map, Mod, Acc) ->
+    mod(Opts, Tag, Address, Map, Mod, Acc, M);
+mod([{T, _} = Opt|Opts], Tag, _Address, Map, Mod, Acc)
+  when T =:= ip; T =:= ifaddr->
+    mod(Opts, Tag, Opt, Map, Mod, [Opt|Acc]);
+mod([Family|Opts], Tag, Address, Map, Mod, Acc) when is_atom(Family) ->
+    case Map of
+	#{Family := M} ->
+	    mod(Opts, Tag, Address, Map, Mod, Acc, M);
+	#{} ->
+	    mod(Opts, Tag, Address, Map, Mod, [Family|Acc])
+    end;
+mod([Opt|Opts], Tag, Address, Map, Mod, Acc) ->
+    mod(Opts, Tag, Address, Map, Mod, [Opt|Acc]);
+mod([], Tag, Address, Map, undefined, Acc) ->
+    {case Address of
+	 {_, {local, _}} ->
+	     case Map of
+		 #{local := Mod} ->
+		     Mod;
+		 #{} ->
+		     inet_db:Tag()
+	     end;
+	 {_, IP} when tuple_size(IP) =:= 8 ->
+	     #{inet := IPv4Mod} = Map,
+	     %% Get the mod, but IPv6 address overrides default IPv4
+	     case inet_db:Tag() of
+		 IPv4Mod ->
+		     #{inet6 := IPv6Mod} = Map,
+		     IPv6Mod;
+		 Mod ->
+		     Mod
+	     end;
+	 _ ->
+	     inet_db:Tag()
+     end, lists:reverse(Acc)};
+mod([], _Tag, _Address, _Map, Mod, Acc) ->
+    {Mod, lists:reverse(Acc)}.
+%%
+mod(Opts, Tag, Address, Map, undefined, Acc, M) ->
+    mod(Opts, Tag, Address, Map, M, Acc);
+mod(Opts, Tag, Address, Map, Mod, Acc, _M) ->
+    mod(Opts, Tag, Address, Map, Mod, Acc).
+
 
 getaddrs_tm({A,B,C,D} = IP, Fam, _)  ->
     %% Only "syntactic" validation and check of family.
@@ -1157,9 +1266,7 @@ gethostbyname_string(Name, Type)
 	    inet ->
 		inet_parse:ipv4_address(Name);
 	    inet6 ->
-		%% XXX should we really translate IPv4 addresses here
-		%% even if we do not know if this host can do IPv6?
-		inet_parse:ipv6_address(Name)
+		inet_parse:ipv6strict_address(Name)
 	end of
 	{ok,IP} ->
 	    {ok,make_hostent(Name, [IP], [], Type)};
@@ -1230,7 +1337,17 @@ gethostbyaddr_tm_native(Addr, Timer, Opts) ->
     end.
 
 -spec open(Fd_or_OpenOpts :: integer() | list(),
-	   Addr :: ip_address(),
+	   Addr ::
+	     socket_address() |
+	     {ip_address() | 'any' | 'loopback', % Unofficial
+	      port_number()} |
+	     {inet, % Unofficial
+	      {ip4_address() | 'any' | 'loopback',
+	       port_number()}} |
+	     {inet6, % Unofficial
+	      {ip6_address() | 'any' | 'loopback',
+	       port_number()}} |
+	     undefined, % Internal - no bind()
 	   Port :: port_number(),
 	   Opts :: [socket_setopt()],
 	   Protocol :: socket_protocol(),
@@ -1249,13 +1366,12 @@ open(FdO, Addr, Port, Opts, Protocol, Family, Type, Module)
     case prim_inet:open(Protocol, Family, Type, OpenOpts) of
 	{ok,S} ->
 	    case prim_inet:setopts(S, Opts) of
+		ok when Addr =:= undefined ->
+		    inet_db:register_socket(S, Module),
+		    {ok,S};
 		ok ->
-		    case if is_list(Addr) ->
-				 bindx(S, Addr, Port);
-			    true ->
-				 prim_inet:bind(S, Addr, Port)
-			 end of
-			{ok, _} -> 
+		    case bind(S, Addr, Port) of
+			{ok, _} ->
 			    inet_db:register_socket(S, Module),
 			    {ok,S};
 			Error  ->
@@ -1273,6 +1389,11 @@ open(Fd, Addr, Port, Opts, Protocol, Family, Type, Module)
   when is_integer(Fd) ->
     fdopen(Fd, Addr, Port, Opts, Protocol, Family, Type, Module).
 
+bind(S, Addr, Port) when is_list(Addr) ->
+    bindx(S, Addr, Port);
+bind(S, Addr, Port) ->
+    prim_inet:bind(S, Addr, Port).
+
 bindx(S, [Addr], Port0) ->
     {IP, Port} = set_bindx_port(Addr, Port0),
     prim_inet:bind(S, IP, Port);
@@ -1313,34 +1434,36 @@ fdopen(Fd, Opts, Protocol, Family, Type, Module) ->
     fdopen(Fd, any, 0, Opts, Protocol, Family, Type, Module).
 
 fdopen(Fd, Addr, Port, Opts, Protocol, Family, Type, Module) ->
-    IsAnyAddr = (Addr == {0,0,0,0} orelse Addr == {0,0,0,0,0,0,0,0} 
-                 orelse Addr == any),
-    Bound = Port == 0 andalso IsAnyAddr,
+    Bound =
+	%% We do not do any binding if default port+addr options
+	%% were given, in order to keep backwards compatability
+	%% with pre Erlang/OTP 17
+	case Addr of
+	    {0,0,0,0} when Port =:= 0 -> true;
+	    {0,0,0,0,0,0,0,0} when Port =:= 0 -> true;
+	    any when Port =:= 0 -> true;
+	    _ -> false
+	end,
     case prim_inet:fdopen(Protocol, Family, Type, Fd, Bound) of
 	{ok, S} ->
 	    case prim_inet:setopts(S, Opts) of
+		ok
+		  when Addr =:= undefined;
+		       Bound ->
+		    inet_db:register_socket(S, Module),
+		    {ok, S};
 		ok ->
-                    case if
-                             Bound ->
-                                 %% We do not do any binding if default
-                                 %% port+addr options where given in order
-                                 %% to keep backwards compatability with
-                                 %% pre Erlang/TOP 17
-                                 {ok, ok};
-                             is_list(Addr) ->
-                                 bindx(S, Addr, Port);
-                             true ->
-                                 prim_inet:bind(S, Addr, Port)
-                         end of
-                        {ok, _} ->
-                            inet_db:register_socket(S, Module),
-                            {ok, S};
-                        Error  ->
-                            prim_inet:close(S),
-                            Error
+		    case bind(S, Addr, Port) of
+			{ok, _} ->
+			    inet_db:register_socket(S, Module),
+			    {ok, S};
+			Error  ->
+			    prim_inet:close(S),
+			    Error
                     end;
 		Error ->
-		    prim_inet:close(S), Error
+		    prim_inet:close(S),
+		    Error
 	    end;
 	Error -> Error
     end.
@@ -1349,11 +1472,14 @@ fdopen(Fd, Addr, Port, Opts, Protocol, Family, Type, Module) ->
 %%  socket stat
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
+-spec i() -> ok.
 i() -> i(tcp), i(udp), i(sctp).
 
+-spec i(socket_protocol()) -> ok.
 i(Proto) -> i(Proto, [port, module, recv, sent, owner,
 		      local_address, foreign_address, state, type]).
 
+-spec i(socket_protocol(), [atom()]) -> ok.
 i(tcp, Fs) ->
     ii(tcp_sockets(), Fs, tcp);
 i(udp, Fs) ->
